/************************************************************/
/** Ce fichier contient des fonctions utilisees pour       **/
/** gere les cameras USB.                                  **/
/************************************************************/

/**** Fichiers d'inclusion ****/

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <linux/videodev2.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#include "libcamformats.h"
#include "libcam.h"

/**** Constantes ****/

#define RGB24_TAILLE	3

/**** Variables globales *****/

struct palette palettes[]={
  {V4L2_PIX_FMT_BGR24,traite_rgb24,24,3},
  {V4L2_PIX_FMT_BGR32,traite_rgb32,32,4},
  {V4L2_PIX_FMT_YUYV,traite_yuyv,24,2},
  {V4L2_PIX_FMT_RGB565,traite_rgb565,16,2},
  {V4L2_PIX_FMT_RGB555,traite_rgb555,15,2},
  {V4L2_PIX_FMT_NV12,traite_yuv420,24,1.5},
  {V4L2_PIX_FMT_SGBRG8,traite_gbrg,8,1},
  {-1,NULL,-1,0}, 
  };

/**** Fonctions de gestion des cameras ****/

/** Initialisation de la camera **/

int camOuvrir(webcam_technique *params){
struct v4l2_capability cap;
struct v4l2_cropcap cropcap;
struct v4l2_crop crop;
struct v4l2_format fmt;
int fd;
int ret;
struct palette *pal;

/* Ouverture du peripherique special correspondant */
#ifdef DEBUG
fprintf(stderr, "Ouverture de la camra '%s'\n",params->periph);
#endif
fd=open(params->periph,O_RDWR);
if(fd<0){
#ifdef DEBUG
  fprintf(stderr,"Erreur systme : camra non prsente (%s)\n",
                 strerror(errno));
#endif
  return -1;
  }
	
/* Recuperation des caracteristiques de la camera */
ret=ioctl(fd,VIDIOC_QUERYCAP,&cap);
if(ret!=0){
#ifdef DEBUG
  fprintf(stderr,"Erreur systeme : caractristiques de la camra inconnues\n");
#endif
erreur:
  close(fd);
  return -1;
  }

/* Verification de la version du pilote */
if(!(cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)){
#ifdef DEBUG
  fprintf(stderr,"Pilote de camera non compatible V4L2\n");
#endif
  goto erreur;
  }
if(!(cap.capabilities & V4L2_CAP_STREAMING)){
#ifdef DEBUG
  fprintf(stderr,"La camera ne permet pas l'alignement memoire\n");
#endif
  goto erreur;
  }

/* Configuration de la camera */
memset(&cropcap,0,sizeof(cropcap));
cropcap.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
ret=ioctl(fd,VIDIOC_CROPCAP,&cropcap);
if(ret==0){
  crop.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
  crop.c=cropcap.defrect;
  ioctl(fd,VIDIOC_S_CROP,&crop);
  }
for(pal=palettes;pal->val>=0;pal++){
  memset(&fmt,0,sizeof(fmt));
  fmt.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
  fmt.fmt.pix.width=params->largeur; 
  fmt.fmt.pix.height=params->hauteur;
  fmt.fmt.pix.pixelformat=pal->val;
  ret=ioctl(fd,VIDIOC_S_FMT,&fmt);
  if(ret==0) break;
  }
for(pal=palettes;pal->val>=0;pal++)
  if(pal->val==fmt.fmt.pix.pixelformat) break;
if(pal->val>=0) goto paltrouve;
#ifdef DEBUG
fprintf(stderr,"Pas de palette connue pour cette camra (%x)\n",
               fmt.fmt.pix.pixelformat);
#endif
goto erreur;
	
/* Initialisation de la capture d'images */
paltrouve:

params->largeur=fmt.fmt.pix.width;
params->hauteur=fmt.fmt.pix.height;
params->palette=pal->val;
params->desc=fd;
#ifdef DEBUG
fprintf(stderr,"Taille d'image finale %dx%d\n",params->largeur,params->hauteur);
fprintf(stderr,"Palette finale %c%c%c%c\n",
                 (pal->val&0xff000000)>>24,(pal->val&0x00ff0000)>>16,
                 (pal->val&0x0000ff00)>>8,(pal->val&0x000000ff));
#endif

/* Alignement memoire */
struct v4l2_requestbuffers req;
struct v4l2_buffer buf;
memset(&req,0,sizeof(req));
req.count=1;
req.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
req.memory=V4L2_MEMORY_MMAP;
ret=ioctl(fd,VIDIOC_REQBUFS,&req);
if(ret<0){
#ifdef DEBUG
  fprintf(stderr,"Erreur systme : alignement mmoire impossible (%s)\n",
                   strerror(errno));
#endif
  goto erreur;
  }
memset(&buf,0,sizeof(buf));
buf.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
buf.memory=V4L2_MEMORY_MMAP;
buf.index=0;
ret=ioctl(fd,VIDIOC_QUERYBUF,&buf);
if(ret<0){
#ifdef DEBUG
  fprintf(stderr,"Erreur systme : alignement mmoire impossible (%s)\n",
                   strerror(errno));
#endif
  goto erreur;
  }
params->images_taille=buf.length;
params->images=mmap(NULL,params->images_taille,PROT_READ|PROT_WRITE,
                    MAP_SHARED,fd,buf.m.offset);
if(params->images==MAP_FAILED){
#ifdef DEBUG
  fprintf(stderr,"Erreur systme : echec de l'alignement mmoire (%s)\n",
                 strerror(errno));
#endif
  goto erreur;
  }
params->rgb_taille=0;
params->rgb=NULL;
params->conversion=0;
	
#ifdef DEBUG
fprintf(stderr, "Camra initialise avec succs\n");
#endif
return 0;
}

int camDemarrer(webcam_technique *params){
int ret;
enum v4l2_buf_type type;
struct v4l2_buffer buf;

memset(&buf,0,sizeof(buf));
buf.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
buf.memory=V4L2_MEMORY_MMAP;
buf.index=0;
ret=ioctl(params->desc,VIDIOC_QBUF,&buf);
if(ret<0){
#ifdef DEBUG
  fprintf(stderr,"Echec d'initialisation du tampon (%s)\n",strerror(errno));
#endif
  return -1;
  }
type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
ret=ioctl(params->desc,VIDIOC_STREAMON,&type);
if(ret<0){
#ifdef DEBUG
  fprintf(stderr,"Impossible de dmarrer la camra (%s)\n",strerror(errno));
#endif
  return -1;
  }
return 0;
}

int camArreter(webcam_technique *params){
int ret;
enum v4l2_buf_type type;

type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
ret=ioctl(params->desc,VIDIOC_STREAMOFF,&type);
if(ret<0){
#ifdef DEBUG
  fprintf(stderr,"Impossible d'arrter la camra (%s)\n",strerror(errno));
#endif
  return -1;
  }
return 0;
}

int camCapturer(webcam_technique *params){
int ret;
struct v4l2_buffer buf;
fd_set fds;
struct timeval tv;

/* Attendre la camera */
FD_ZERO(&fds); FD_SET(params->desc,&fds);
tv.tv_sec=10; tv.tv_usec=0;
do ret=select(params->desc+1,&fds,NULL,NULL,&tv); while(ret<0 && EINTR==errno);
if(ret<0){
#ifdef DEBUG
  fprintf(stderr,"Problme d'attente sur descripteur (%s)\n",strerror(errno));
#endif
  exit(-1);
  }
if(ret==0){
#ifdef DEBUG
  fprintf(stderr,"Camra trop lente\n");
#endif
  exit(-1);
  }

/* Recuperation du tampon */ 
memset(&buf,0,sizeof(buf));
buf.type=V4L2_BUF_TYPE_VIDEO_CAPTURE;
buf.memory=V4L2_MEMORY_MMAP;
ret=ioctl(params->desc,VIDIOC_DQBUF,&buf);
if(ret<0){
#ifdef DEBUG
  fprintf(stderr,"Echec de la capture d'image (%s)\n",strerror(errno));
#endif
  return -1;
  }
ret=ioctl(params->desc,VIDIOC_QBUF,&buf);
if(ret<0){
#ifdef DEBUG
  fprintf(stderr,"Echec de la synchronisation (%s)\n",strerror(errno));
#endif
  return -1;
  }
params->conversion=0;
return 0;
}

/** Convertir l'image de la camera au format RGB24 **/

int camImageRGBConversion(webcam_technique *params){
params->rgb_taille=RGB24_TAILLE*params->hauteur*params->largeur;
if(params->rgb==NULL){
#ifdef DEBUG
  fprintf(stderr,"Allocation d'espace pour l'image RGB24 ...");
#endif
  params->rgb=(void *)malloc(params->rgb_taille);
#ifdef DEBUG
  fprintf(stderr,"Fait\n");
#endif
  }
if(!params->conversion){
  int pal=params->palette;
  int i=0;
  while(palettes[i].val>0 && palettes[i].val!=pal) i++;
  if(palettes[i].val>0){
#ifdef DEBUG
    fprintf(stderr,"Conversion de l'image en RGB24 ...");
#endif
    palettes[i].rgb(
      params->rgb,params->images,
      params->largeur,params->hauteur,
      params->rgb_taille);
#ifdef DEBUG
    fprintf(stderr,"Fait\n");
#endif
    }
  params->conversion=1;
  }
return 0;
}

/** Relacher la camera **/

int camFermer(webcam_technique *params){
#ifdef DEBUG
  fprintf(stderr,"Libration de camra (%s) ...",params->periph);
#endif
munmap(params->images,params->images_taille);
close(params->desc);
#ifdef DEBUG
  fprintf(stderr,"Fait.\n");
#endif

return 0;
}

/** Gerer la luminosite **/

int camLuminosite(webcam_technique *params,int value){
struct v4l2_queryctrl queryctrl;
struct v4l2_control control;
int ret;

memset(&queryctrl,0,sizeof (queryctrl));
queryctrl.id=V4L2_CID_BRIGHTNESS;
ret=ioctl(params->desc,VIDIOC_QUERYCTRL,&queryctrl);
if(ret<0 || (queryctrl.flags&V4L2_CTRL_FLAG_DISABLED)){
#ifdef DEBUG
  fprintf(stderr,"Pas de contrle de luminosit\n");
#endif
  return -1;
  }
memset(&control,0,sizeof(control));
control.id=V4L2_CID_BRIGHTNESS;
control.value=queryctrl.minimum+(queryctrl.maximum-queryctrl.minimum)*value/100;
ret=ioctl(params->desc,VIDIOC_S_CTRL,&control);
if(ret<0){
#ifdef DEBUG
  fprintf(stderr,"Echec de la modification de luminosit\n");
#endif
  return -1;
  }
return 0;
}


