/******************************************************************/
/** Ce fichier definit les fonctions pour la conversions des     **/
/** formats d'images.                                            **/
/******************************************************************/

/**** Les fichiers d'inclusion ****/

#include <string.h>

/**** Les fonctions de conversion ****/

static
unsigned char
citb(int i)
{
  if (i >= 256)
    return 255;
  if (i < 0)
    return 0;
  return i;
}

void
traite_yuv420p(unsigned char *dst, unsigned char *src, int x, int y, int size)
{
  const unsigned char *u, *v, *bu, *bv;
  unsigned int dx, dy, uvc, buvc;
  
  u = src + x * y;
  v = u + (x / 2) * (y / 2);
  
  buvc = 0;
  for (dy = 0; dy < y; dy++)
  {
    bu = u;
    bv = v;
    uvc = 0;
    
    for (dx = 0; dx < x; dx++)
    {
      *dst++ = citb((int) *src + 1.140 * ((int) *v - 128));
      *dst++ = citb((int) *src - 0.396 * ((int) *u - 128) - 0.581 * ((int) *v - 128));
      *dst++ = citb((int) *src++ + 2.029 * ((int) *u - 128));
      
      uvc++;
      if (uvc >= 2)
      {
        uvc = 0;
        u++;
        v++;
      }
    }

    buvc++;
    if (buvc < 2)
    {
      u = bu;
      v = bv;
    }
    else
      buvc = 0;
  }
}

void
traite_rgb24(unsigned char *dst, unsigned char *src, int x, int y, int size)
{
  memcpy(dst, src, size);
}

void
traite_rgb32(unsigned char *dst, unsigned char *src, int x, int y, int size)
{
  unsigned char *dstend;
  
  dstend = dst + size; 
  while (dst < dstend)
  {
    *dst++ = *src++;
    *dst++ = *src++;
    *dst++ = *src++;
    src++;
  }
}

void
traite_yuyv(unsigned char *dst, unsigned char *src, int x, int y, int size)
{
  unsigned char *dstend;
  unsigned char u, v;
  int i;
  
  dstend = dst + size;
  
  while (dst < dstend)
  {
    u = src[1];
    v = src[3];

    for (i = 0; i < 2; i++)
    {
      *dst++ = citb((int) *src + 1.140 * ((int) v - 128));
      *dst++ = citb((int) *src - 0.396 * ((int) u - 128) - 0.581 * ((int) v - 128));
      *dst++ = citb((int) *src + 2.029 * ((int) u - 128));
      src += 2;
    }
  }
}

void
traite_yuv420(unsigned char *dst, unsigned char *src, int x, int y, int size)
{
  unsigned char *dstend;
  unsigned char *u, *v;
  int ypl, bpl;
  int i, j, k, l;

  dstend = dst + size;
  
  u = src + 4;
  bpl = x * 1.5;
  v = u + bpl;
  ypl = x / 4;
  
  /* YYYY UU YYYY UU
   * YYYY VV YYYY VV */
  while (dst < dstend)
  {
    for (l = 0; l < 2; l++)
    {
      for (k = 0; k < ypl; k++)
      {
        for (j = 0; j < 2; j++)
        {
          for (i = 0; i < 2; i++)
          {
            *dst++ = citb((int) *src + 1.140 * ((int) *v - 128));
            *dst++ = citb((int) *src - 0.396 * ((int) *u - 128) - 0.581 * ((int) *v - 128));
            *dst++ = citb((int) *src + 2.029 * ((int) *u - 128));
            src++;
          }
          u++;
          v++;
        }
        src += 2;
        u += 4;
        v += 4;
      }
      u -= bpl;
      v -= bpl;
    }
    u += bpl * 2;
    v += bpl * 2;
  }
}

void
traite_rgb565(unsigned char *dst, unsigned char *src, int x, int y, int size)
{
  unsigned char *dstend;

  dstend = dst + size;

  while (dst < dstend)
  {
    /* rrrrrggg gggbbbbb -> gggbbbbb rrrrrggg */
    *dst++ = (src[1] & 0xf8) | ((src[1] >> 3) & 0x07);
    *dst++ = (src[1] << 5) | ((src[0] >> 3) & 0x1c) | ((src[0] >> 5) & 0x03);
    *dst++ = (src[0] << 3) | (src[0] & 0x07);
    src += 2;
  }
}

void
traite_rgb555(unsigned char *dst, unsigned char *src, int x, int y, int size)
{
  unsigned char *dstend;

  dstend = dst + size;

  while (dst < dstend)
  {
    /* 0rrrrrgg gggbbbbb -> gggbbbbb 0rrrrrgg */
    *dst++ = ((src[1] << 1) & 0xf8) | ((src[1] >> 2) & 0x07);
    *dst++ = (src[1] << 6) | ((src[0] >> 2) & 0x38) | ((src[0] >> 5) & 0x07);
    *dst++ = (src[0] << 3) | (src[0] & 0x07);
    src += 2;
  }
}

#define bayer_R(tab,x,y,width) (tab[0 + 3 * ((x) + width * (y))])
#define bayer_G(tab,x,y,width) (tab[1 + 3 * ((x) + width * (y))])
#define bayer_B(tab,x,y,width) (tab[2 + 3 * ((x) + width * (y))])

#define bayer_RGB(tab,x,y,width) (tab[(x) + width * (y)])

static void bayer_copy(unsigned char *src, unsigned char *dst, int x, int y, int width)
{

  bayer_G(dst,x + 0, y + 0, width) = bayer_RGB(src, x + 0, y + 0, width);
  bayer_G(dst,x + 1, y + 1, width) = bayer_RGB(src, x + 1, y + 1, width);
  bayer_G(dst,x + 0, y + 1, width) = bayer_G(dst,x + 1, y + 0, width) = ((unsigned long)bayer_RGB(src, x + 0, y + 0, width) + (unsigned long)bayer_RGB(src, x + 1, y + 1, width)) / 2;
  bayer_R(dst,x + 0, y + 0, width) = bayer_R(dst,x + 1, y + 0, width) = bayer_R(dst,x + 1, y + 1, width) = bayer_R(dst,x + 0, y + 1, width) = bayer_RGB(src, x + 0, y + 1, width);
  bayer_B(dst,x + 1, y + 1, width) = bayer_B(dst,x + 0, y + 0, width) = bayer_B(dst,x + 0, y + 1, width) = bayer_B(dst,x + 1, y + 0, width) = bayer_RGB(src, x + 1, y + 0, width);
}

static void bayer_bilinear(unsigned char *src, unsigned char *dst, int x, int y, int width)
{
  bayer_R(dst,x + 0, y + 0, width) = ((unsigned long)bayer_RGB(src,x + 0, y + 1, width) + (unsigned long)bayer_RGB(src,x + 0, y - 1, width)) / 2;
  bayer_G(dst,x + 0, y + 0, width) = bayer_RGB(src,x + 0, y + 0, width);
  bayer_B(dst,x + 0, y + 0, width) = ((unsigned long)bayer_RGB(src,x - 1, y + 0, width) + (unsigned long)bayer_RGB(src,x + 1, y + 0, width)) / 2;

  bayer_R(dst,x + 0, y + 1, width) = bayer_RGB(src,x + 0, y + 1, width);
  bayer_G(dst,x + 0, y + 1, width) = ((unsigned long)bayer_RGB(src,x + 0, y + 0, width) + (unsigned long)bayer_RGB(src,x + 0, y + 2, width)
       + (unsigned long)bayer_RGB(src,x - 1, y + 1, width) + (unsigned long)bayer_RGB(src,x + 1, y + 1, width)) / 4;
  bayer_B(dst,x + 0, y + 1, width) = ((unsigned long)bayer_RGB(src,x + 1, y + 0, width) + (unsigned long)bayer_RGB(src,x - 1, y + 0, width)
       + (unsigned long)bayer_RGB(src,x + 1, y + 2, width) + (unsigned long)bayer_RGB(src,x - 1, y + 2, width)) / 4;

  bayer_R(dst,x + 1, y + 0, width) = ((unsigned long)bayer_RGB(src,x + 0, y + 1, width) + (unsigned long)bayer_RGB(src,x + 2, y + 1, width)
       + (unsigned long)bayer_RGB(src,x + 0, y - 1, width) + (unsigned long)bayer_RGB(src,x + 2, y - 1, width)) / 4;
  bayer_G(dst,x + 1, y + 0, width) = ((unsigned long)bayer_RGB(src,x + 0, y + 0, width) + (unsigned long)bayer_RGB(src,x + 2, y + 0, width)
       + (unsigned long)bayer_RGB(src,x + 1, y - 1, width) + (unsigned long)bayer_RGB(src,x + 1, y + 1, width)) / 4;
  bayer_B(dst,x + 1, y + 0, width) = bayer_RGB(src,x + 1, y + 0, width);

  bayer_R(dst,x + 1, y + 1, width) = ((unsigned long)bayer_RGB(src,x + 0, y + 1, width) + (unsigned long)bayer_RGB(src,x + 2, y + 1, width)) / 2;
  bayer_G(dst,x + 1, y + 1, width) = bayer_RGB(src,x + 1, y + 1, width);
  bayer_B(dst,x + 1, y + 1, width) = ((unsigned long)bayer_RGB(src,x + 1, y + 0, width) + (unsigned long)bayer_RGB(src,x + 1, y + 2, width)) / 2;
}


void
traite_gbrg(unsigned char *dst, unsigned char *src, int x, int y, int size)
{
  int i, j;

  for (i = 0; i < x; i += 2)
    for (j = 0; j < y; j += 2)
      if (i == 0 || j == 0 || i == x - 2 || j == y - 2)
        bayer_copy(src, dst, i, j, x);
      else
        bayer_bilinear(src, dst, i, j, x);
}

