/*****************************************************************/
/*  Executable permettant de recuperer des cliches de la webcam  */
/*****************************************************************/

/** Fichiers d'inclusion **/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
#include <signal.h>

#include "../Camera/libcam.h"
#include "../IPC/libipc.h"

/** Constantes **/

#define QUEUE_CAPTURE	0xffbb

#define IMAGE_LARGEUR	1024
#define IMAGE_HAUTEUR	768

#define	PNG_OUTIL	"pnmtopng"

#define MODE_SERVEUR	0
#define MODE_CLIENT	1

#define	FAUX		0
#define	VRAI		1

#define MAX_CHAINE	1024

#define COMMANDE_PHOTO	"photo"

/** Variables globales **/

webcam_technique camera;

/** Initialise la webcam **/

webcam_technique initcam(char *periph){
webcam_technique cam;
int status;
strcpy(cam.periph,periph);
cam.hauteur=IMAGE_LARGEUR;
cam.largeur=IMAGE_HAUTEUR;
status=camOuvrir(&cam);
if(status<0){
  fprintf(stderr,"Impossible d'ouvrir la camera !\n");
  exit(-1);
  }
status=camLuminosite(&cam,90);
if(status<0){
  fprintf(stderr,"Impossible de regler la luminosite !\n");
  }
status=camDemarrer(&cam);
if(status<0){
  fprintf(stderr,"Impossible de demarrer la camera !\n");
  exit(-1);
  }
return cam;
}

/** Generer l'image en format PNG **/

void creerPNG(unsigned char *data,int x,int y)
{
FILE *f=popen(PNG_OUTIL,"w");
fprintf(f,"P6\n");
fprintf(f,"%d %d\n",x,y);
fprintf(f,"255\n");
fwrite(data,x*y*3,1,f);
pclose(f);	
}

/** Arret general **/

void terminer(int signal){
int status;
status=camArreter(&camera);
if(status<0){
  fprintf(stderr,"Impossible d'arreter la camera !\n");
  exit(-1);
  }
status=camFermer(&camera);
if(status<0){
  fprintf(stderr,"Impossible de fermer la camera !\n");
  exit(-1);
  }
destructionQueue();
}

/** Afficher la syntaxe **/

void usage(char *nom){
fprintf(stderr,"Usage:\n");
fprintf(stderr,"  %s -s <periph>\n",nom);
fprintf(stderr,"  %s -c\n",nom);
exit(-1);
}

/** Fonction principale **/

int main(int argc,char *argv[]){

// Lecture des parametres
char *periph="/dev/video0";
int mode=-1;
int flag;
static struct option long_options[]={
  {"server",0,0,'s'},
  {"client",0,0,'c'},
  {0,0,0,0}
  };
while(1){
  flag=getopt_long(argc,argv,"sc",long_options,NULL);
  if(flag==-1) break;
  switch(flag){
    case 's': mode=MODE_SERVEUR; break;
    case 'c': mode=MODE_CLIENT; break;
    default: usage(argv[0]); break;
    }
  }
if(mode<0) usage(argv[0]);
if(mode==MODE_SERVEUR){
  if(argc>optind) periph=argv[optind];
  if(argc>optind+1) usage(argv[0]);
  }
if(mode==MODE_CLIENT){
  if(argc>optind) usage(argv[0]);
  }
 
// Mode serveur
if(mode==MODE_SERVEUR){
  signal(SIGINT,terminer);
  initialisationQueue(QUEUE_CAPTURE,VRAI);
  camera=initcam(periph);
  while(1){
    char commande[MAX_CHAINE];
    int taille=MAX_CHAINE;
    int status;
    status=camCapturer(&camera);
    if(status<0){
      fprintf(stderr,"Impossible de capturer une image !\n");
      return -1;
      }
    status=camImageRGBConversion(&camera);
    if(status<0){
      fprintf(stderr,"Impossible de convertir l'image !\n");
      return -1;
      }
    recevoirMessage(QUEUES_COMMANDE_TYPE,commande,&taille,FAUX);
    if(taille>0){
      envoyerBinaire((char *)&camera.largeur,sizeof(int));
      envoyerBinaire((char *)&camera.hauteur,sizeof(int));
      envoyerBinaire((char *)camera.rgb,3*camera.largeur*camera.hauteur*sizeof(char));
      }
    sleep(1);
    }
  }

// Mode client
if(mode==MODE_CLIENT){
  int largeur,hauteur;
  unsigned char *image=NULL;
  initialisationQueue(QUEUE_CAPTURE,FAUX);
  envoyerMessage(QUEUES_COMMANDE_TYPE,COMMANDE_PHOTO,strlen(COMMANDE_PHOTO)+1);
  recevoirBinaire((char *)&largeur,sizeof(int));
  recevoirBinaire((char *)&hauteur,sizeof(int));
  if(largeur>0 && hauteur>0){
    int taille=3*largeur*hauteur*sizeof(char);
    image=(unsigned char *)malloc(taille);
    recevoirBinaire((char *)image,taille);
    creerPNG(image,largeur,hauteur);
    }
  }

return 0;
}
