/** fichier libipc.c **/

/********************************************************/
/** Ce fichier centralise toutes les actions au  sujet **/
/** des IPCs (ici uniquement des queues de messages)   **/
/********************************************************/

/** Fichiers d'inclusion **/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>

#include "libipc.h"

/** Constantes privees **/

#define TAILLE_BLOC_BINAIRE	6*1024

/** Variables globales **/

static int queue;	/* Identificateur de la queue de messages */

/** Fonction d'initialisation de la queue de commandes **/

void initialisationQueue(int clef,unsigned char creation){
queue=msgget(clef,QUEUES_PERMISSIONS|(creation?IPC_CREAT:0));
if(queue<0){
    perror("creationQueueCommande.msgget");
    exit(-1);
    }
#ifdef DEBUG
fprintf(stderr,"Identifieur de la queue de commandes : %x\n",queueCommande);
#endif
}

/** Destruction de la queue de messages **/

void destructionQueue(void){ msgctl(queue,IPC_RMID,0); }

/** Envoi d'un message generique par queue IPC **/

void envoyerMessage(int type,void *message,int taille){
struct msgbuf *temp;

temp=(struct msgbuf *)malloc(sizeof(struct msgbuf)+taille);
temp->mtype=type;
memcpy((void *)temp->mtext,(void *)message,taille);
if(msgsnd(queue,temp,taille,0)<0){
    perror("envoyerMessage.msgsnd");
    exit(-1);
    }           
free(temp);
}

/** Reception d'un message generique par queue IPC **/

void recevoirMessage(int type,void *message,int *taille,unsigned char bloquer){
struct msgbuf *temp=(struct msgbuf *)malloc(sizeof(struct msgbuf)+*taille);
int drapeau=bloquer?0:IPC_NOWAIT;
int octets=msgrcv(queue,temp,*taille,type,drapeau);
if(octets<0 && errno==ENOMSG){ *taille=0; free(temp); return; }
if(octets<0){
    perror("recevoirMessage.msgrcv");
    exit(-1);
    }
if(octets<*taille) *taille=octets;
memcpy(message,(void *)temp->mtext,*taille);
free(temp);
}

/** Envoi et reception de donnees binaires **/

void envoyerBinaire(char *donnees,int taille){
#ifdef DEBUG
fprintf(stderr,"Envoyer binaire (taille %d)\n",taille);
#endif
while(taille>0){
  int taille_bloc;
  taille_bloc=(taille<TAILLE_BLOC_BINAIRE)?taille:TAILLE_BLOC_BINAIRE;
  envoyerMessage(QUEUES_BINAIRE_TYPE,(void *)donnees,taille_bloc);
  taille -= taille_bloc; donnees += taille_bloc;
  }
}

void recevoirBinaire(char *donnees,int taille){
#ifdef DEBUG
fprintf(stderr,"Reception binaire (taille maximum %d) ... ",taille);
#endif
while(taille>0){
  int taille_bloc;
  taille_bloc=(taille<TAILLE_BLOC_BINAIRE)?taille:TAILLE_BLOC_BINAIRE;
  recevoirMessage(QUEUES_BINAIRE_TYPE,(void *)donnees,&taille_bloc,1);
  taille -= taille_bloc; donnees += taille_bloc;
  }
#ifdef DEBUG
fprintf(stderr,"Fait\n");
#endif
}
